unit MediaFormat;

{$IMPORTEDDATA OFF}

interface

uses Windows, WinAdd, NTypes, NLib;

type
  TWaveFormatExFFMPEG = record
    nCodecId: Integer;
    wfex: TWaveFormatEx;
  end;
  PWaveFormatExFFMPEG = ^TWaveFormatExFFMPEG;

  TVorbisFormat2 = record
    Channels: DWORD;
    SamplesPerSec: DWORD;
    BitsPerSample: DWORD;
    HeaderSize: array [0..2] of DWORD;
  end;
  PVorbisFormat2 = ^TVorbisFormat2;

const
  FORMAT_WaveFormatExFFMPEG: TGUID = '{35189950-CAC9-4C8D-819D-B6FAEE15DD9D}';
  FORMAT_VorbisFormat2: TGUID = '{B36E107F-A938-4387-93C7-55E966757473}';

function ContainerName(const F: TStream): string;
function MediaSubTypeName(const MediaSubType: TGUID): string;
function WaveFormatName(WaveFormatTag: Word): string;

implementation

type
  TGuidName = record
    Guid: TGUID;
    Name: string;
  end;

  TTagName = record
    Tag: Cardinal;
    Name: string;
  end;

const
  // AAC-ADTS LAVF/LAVC specific
  S_MEDIASUBTYPE_AAC_ADTS = '{53544441-0000-0010-8000-00AA00389B71}';

  MediaSubTypeNames: array [0..0] of TGUIDName = (
    (Guid: S_MEDIASUBTYPE_AAC_ADTS; Name: 'AAC')
  );

  WaveFormatTagNames: array [0..20] of TTagName = (
    (Tag: WAVE_FORMAT_PCM; Name: 'PCM'),
    (Tag: WAVE_FORMAT_ADPCM; Name: 'ADPCM'),
    (Tag: WAVE_FORMAT_WMAVOICE9; Name: 'WMA9Voice'),
    (Tag: WAVE_FORMAT_MPEG; Name: 'MPEG'),
    (Tag: WAVE_FORMAT_MPEGLAYER3; Name: 'MP3'),
    (Tag: WAVE_FORMAT_DOLBY_AC3_SPDIF; Name: 'AC3'),
    (Tag: WAVE_FORMAT_RAW_AAC1; Name: 'AAC'),
    (Tag: WAVE_FORMAT_MSAUDIO1; Name: 'MSAUDIO1'),
    (Tag: WAVE_FORMAT_WMAUDIO2; Name: 'WMA'),
    (Tag: WAVE_FORMAT_WMAUDIO3; Name: 'WMA9Pro'),
    (Tag: WAVE_FORMAT_WMAUDIO_LOSSLESS; Name: 'WMA9Lossless'),
    (Tag: WAVE_FORMAT_MPEG_ADTS_AAC; Name: 'ADTSAAC'),
    (Tag: WAVE_FORMAT_DVM; Name: 'AC-3'),
    (Tag: WAVE_FORMAT_OGG_VORBIS_MODE_1; Name: 'VORBIS1'),
    (Tag: WAVE_FORMAT_OGG_VORBIS_MODE_2; Name: 'VORBIS2'),
    (Tag: WAVE_FORMAT_OGG_VORBIS_MODE_3; Name: 'VORBIS3'),
    (Tag: WAVE_FORMAT_OGG_VORBIS_MODE_1_PLUS; Name: 'VORBIS1P'),
    (Tag: WAVE_FORMAT_OGG_VORBIS_MODE_2_PLUS; Name: 'VORBIS2P'),
    (Tag: WAVE_FORMAT_OGG_VORBIS_MODE_3_PLUS; Name: 'VORBIS3P'),
    (Tag: WAVE_FORMAT_OPUS; Name: 'OPUS'),
    (Tag: WAVE_FORMAT_FLAC; Name: 'FLAC')
  );

function DetectMPEG(var head: string; F: TStream): string;
type
  TID3v2Header = packed record
    Tag: array [0..2] of Char;
    Version: Byte;
    Revision: Byte;
    Flags: Byte;
    Size: UInt32;
  end;
  PID3v2Header = ^TID3v2Header;

  function DecodeSyncSafeInt(Value: UInt32): UInt32;
  begin
    Result := Value and $7f or Value and $7f00 shr 1 or
        Value and $7f0000 shr 2 or Value and $7f000000 shr 3;
  end;
var
  i: Integer;
  n: Cardinal;
  foundseqheader, hasts: Boolean;
  id3v2: string;
begin
  if Copy(head, 1, 3) = 'ID3' then begin
    with PID3V2Header(@head[1])^ do begin
      id3v2 := Format(' with ID3v2.{}.{}', [Version, Revision]);
      F.Seek(SizeOf(TID3V2Header) + DecodeSyncSafeInt(Be32ToH(Size)));
      F.Read(head[1], Length(head), n);
      SetLength(head, n);
    end;
    hasts := False;
  end else begin
    id3v2 := '';
    hasts := head[1] = #$47;
  end;

  foundseqheader := False;
  i := 1;
  while i <= (Length(head) - 4) do begin
    if id3v2 = '' then begin
      if (head[i] = #$00) and (head[i+1] = #$00) and (head[i+2] = #$01) then begin
        case Ord(head[i+3]) of
          $b3: foundseqheader := True; // sequence header
          $b5: // squence extension
            if foundseqheader then begin
              if hasts then
                Result := 'MPEG-2 Transport Stream'
              else
                Result := 'MPEG-2 Elementary Stream';
              Exit;
            end;
          $b2, // use data
          $b8: // group of pictures
            if foundseqheader then begin
              Result := 'MPEG-1 Elementary Stream';
              Exit;
            end;
          $ba: // pack header
            begin
              if (Ord(head[i+4]) and $c0) = $40 then
                Result := 'MPEG-2 Program Stream'
              else
                Result := 'MPEG-1 System Stream';
              Exit;
            end;
        end;
      end;
    end;

    // audio frame
    if not hasts and not foundseqheader then begin
      if (head[i] = #$ff) and ((Ord(head[i+1]) and $e0) = $e0) then begin
        Result := 'MPEG Audio' + id3v2;
        Exit;
      end;
    end;

    Inc(i);
  end;
  Result := '';
end;

function ContainerName(const F: TStream): string;
const
  ASF_Header_Object: TGUID = '{75B22630-668E-11CF-A6D9-00AA0062CE6C}';
  EBML_ID_HEADER = #$1a#$45#$df#$a3;
var
  i: Integer;
  n: Cardinal;
  buf, fcc, doctype: string;
begin
  SetString(buf, nil, 8192);
  F.Read(buf[1], Length(buf), n);
  SetLength(buf, n);
  fcc := Copy(buf, 1, 4);
  if fcc = 'RIFF' then
    Result := 'RIFF ' +  Copy(Buf, 9, 4)
  else if IsEqualGUID(PGUID(@buf[1])^, ASF_Header_Object) then
    Result := 'ASF' // used by container type check in Player
  else if fcc = EBML_ID_HEADER then begin
    i := Length(EBML_ID_HEADER) + 1;
    while i <= (Length(buf) - 4) do begin
      if (buf[i] = #$42) and (buf[i+1] = #$82) then begin // EBML_ID_DOCTYPE
        doctype := Copy(buf, i+3, Ord(buf[i+2]) and $7f);
        if doctype = 'matroska' then
          Result := 'Matroska'
        else if doctype = 'webm' then
          Result := 'WebM'
        else
          Result := 'EBML unknown doctype';
        Exit;
      end;
      Inc(i);
    end;
    Result := '';
  end else if Copy(buf, 5, 4) = 'ftyp' then
    Result := 'MP4 ' + Copy(buf, 9, 4)
  else if fcc = 'FLV'#1 then
    Result := 'FLV'
  else if fcc = 'OggS' then
    Result := 'Ogg'
  else if fcc = 'fLaC' then
    Result := 'FLAC'
  else
    Result := DetectMPEG(buf, f);
end;

function WaveFormatName(WaveFormatTag: Word): string;
var
  i: Integer;
begin
  for i := 0 to High(WaveFormatTagNames) do
    with WaveFormatTagNames[i] do
      if WaveFormatTag = Tag then begin
        Result := Name;
        Exit;
      end;
  Result := '';
end;

function MediaSubTypeName(const MediaSubType: TGUID): string;
var
  i: Integer;
begin
  for i := 0 to High(MediaSubTypeNames) do
    with MediaSubTypeNames[i] do
      if IsEqualGUID(MediaSubType, Guid) then begin
        Result := Name;
        Exit;
      end;
  Result := '';
end;

end.
